"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var hasOwnProperty_1 = require("./hasOwnProperty");
var path_1 = require("./path");
function deepCompare(actual, expected) {
    var differences = [];
    recursiveDeepCompare(actual, expected, differences);
    differences.sort(compareDifferences);
    return differences;
}
exports.deepCompare = deepCompare;
function recursiveDeepCompare(a, b, differences, path, aStack, bStack) {
    if (path === void 0) { path = ''; }
    if (aStack === void 0) { aStack = []; }
    if (bStack === void 0) { bStack = []; }
    if (Object.is(a, b)) {
        return;
    }
    var aIndex = aStack.indexOf(a);
    var bIndex = bStack.indexOf(b);
    if (aIndex >= 0 || bIndex >= 0) {
        if (aIndex !== bIndex) {
            differences.push({ path: path, message: 'circular reference mismatch' });
        }
        return;
    }
    var type = getType(a);
    if (type !== getType(b)) {
        differences.push({
            path: path,
            message: "expected type " + getType(b) + ", received " + type
        });
        return;
    }
    if (type === 'object') {
        aStack.push(a);
        bStack.push(b);
        var keys = keysOf(a, b);
        for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
            var key = keys_1[_i];
            var keyPath = path_1.joinPath(path, key);
            if (!hasOwnProperty_1.hasOwnProperty(a, key)) {
                differences.push({ path: keyPath, message: "property missing" });
            }
            else if (!hasOwnProperty_1.hasOwnProperty(b, key)) {
                differences.push({ path: keyPath, message: "property should not be present" });
            }
            else {
                recursiveDeepCompare(a[key], b[key], differences, keyPath, aStack, bStack);
            }
        }
        aStack.pop();
        bStack.pop();
        return;
    }
    if (type === 'array') {
        if (a.length !== b.length) {
            differences.push({ path: path, message: "expected length " + b.length + ", received " + a.length });
        }
        aStack.push(a);
        bStack.push(b);
        for (var i = 0; i < Math.max(a.length, b.length); i++) {
            if (i >= a.length) {
                differences.push({ path: path_1.joinPath(path, i), message: "item missing" });
            }
            else if (i >= b.length) {
                differences.push({ path: path_1.joinPath(path, i), message: "item should not be present" });
            }
            else {
                recursiveDeepCompare(a[i], b[i], differences, path_1.joinPath(path, i), aStack, bStack);
            }
        }
        aStack.pop();
        bStack.pop();
        return;
    }
    differences.push({ path: path, message: "expected " + b + ", received " + a });
}
exports.recursiveDeepCompare = recursiveDeepCompare;
function getType(value) {
    if (Array.isArray(value)) {
        return 'array';
    }
    else if (value === null) {
        return 'null';
    }
    else if (value === undefined) {
        return 'undefined';
    }
    else {
        return typeof value;
    }
}
function keysOf(a, b) {
    var keys = [];
    for (var key in a) {
        if (hasOwnProperty_1.hasOwnProperty(a, key) && keys.indexOf(key) === -1) {
            keys.push(key);
        }
    }
    for (var key in b) {
        if (hasOwnProperty_1.hasOwnProperty(b, key) && keys.indexOf(key) === -1) {
            keys.push(key);
        }
    }
    return keys;
}
function compareDifferences(a, b) {
    return lexCompare(a.path, b.path);
}
function lexCompare(a, b) {
    return (a < b ? -1 : (a > b ? 1 : 0));
}
